/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file server/connection.c
 * @brief module responsible for the peer-to-peer connections
 *
 * This file contains the connection table which lists all the current
 * connections of the node with other hosts and buffers outgoing
 * packets to these hosts.  The connection table also contains state
 * information such as sessionkeys, credibility and the last time we
 * had host activity.<p>
 *
 * This code is responsible for exchanging a sessionkey with another
 * peer, grouping several messages into a larger packet, padding with
 * noise, encryption and deferred sending of these messages.<p>
 *
 * The file is organized as follows:
 * 
 * a) includes
 * b) defines
 * c) typedefs
 * d) global variables
 * e) code
 * <p>
 *
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 **/ 

/* ******************* includes ******************** */
#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/statistics.h"
#include "util/storage.h"
#include "util/symcipher.h"
#include "util/tcp_return.h"
#include "util/xmalloc.h"
#include "util/statuscalls.h"
#include "server/knownhosts.h"
#include "server/traffic.h"
#include "server/pingpong.h"
#include "server/handler.h"
#include "server/tcpserver.h"
#include "server/heloexchange.h"

/* **************** defines ************ */

/* tuning parameters */

#define DEBUG_CONNECTION NO

#if DEBUG_CONNECTION == 2
#define ENTRY() LOG(LOG_DEBUG, "Method entry: %s defined in %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
#else
#define ENTRY() ;
#endif

/**
 * If an attempt to establish a connection is not answered
 * within 150s, drop.
 **/
#define SECONDS_NOPINGPONG_DROP 150

/**
 * If an established connection is inactive for 5 minutes, 
 * drop.
 **/
#define SECONDS_INACTIVE_DROP 300

/**
 * After 2 minutes on an inactive connection, probe the other
 * node with a ping if we have achieved less than 50% of our
 * connectivity goal.
 **/
#define SECONDS_PINGATTEMPT 120

/**
 * How big do we estimate should the send buffer be?  (can grow bigger
 * if we have many requests in a short time, but if it is larger than
 * this, we start do discard expired entries.)  Computed as "MTU /
 * querySize" plus a bit with the goal to be able to have at least
 * enough small entries to fill a message completely *and* to have
 * some room to manouver.
 **/
#define TARGET_SBUF_SIZE 40

/**
 * How frequent (per connection!) may we attempt to solve the knapsack
 * problem and send a message out? Note that setting this value higher
 * reduces the CPU overhead while a lower value can improve thoughput.
 *
 * The value is now adjusted according to how fast we perceive the CPU
 * to be...
 **/
unsigned int MAX_SEND_FREQUENCY = 50 * cronMILLIS;

/**
 * Priority for special administrative messages that
 * for example overrules drop-rules.
 **/
#define EXTREME_PRIORITY 0xFFFFFF

/**
 * High priority message that needs to go through fast,
 * but not if policies would be disregarded.
 **/
#define ADMIN_PRIORITY 0xFFFF

/**
 * Expected MTU for a connection (1500 for Ethernet)
 **/
#define EXPECTED_MTU 1500

/**
 * How many MTU size messages to we want to transmit
 * perr SECONDS_INACTIVE_DROP interval? (must be >=4 to
 * keep connection alive with reasonable probability).
 **/
#define TARGET_MSG_SID 128

/* status constants */

#define STAT_DOWN             0
#define STAT_WAITING_FOR_PING 1
#define STAT_WAITING_FOR_PONG 2
#define STAT_UP               3

/* ******************** typedefs ******************* */

/**
 * Type of the linked list of send callbacks (to
 * implement a round-robbin invocation chain).
 **/
typedef struct SendCallbackList__ {
  /**
   * Minimum number of bytes that must be available
   * to call this callback.
   **/
  unsigned int minimumPadding;

  /**
   * The callback method.
   **/
  BufferFillCallback callback;

  /**
   * Did we say that this is a linked list?
   **/
  struct SendCallbackList__ * next;

} SendCallbackList;

/**
 * Record for state maintanance between scanHelperCount,
 * scanHelperSelect and scanForHosts.
 **/
typedef struct {
  unsigned int index;
  unsigned int matchCount;
  int costSelector;  
  HostIdentity match;
} IndexMatch;

typedef struct fENHWrap {
  PerNodeCallback method;
  void * arg;
} fENHWrap;

/* *********** flags for SendEntry.flags ********** */

/* no flags */
#define SE_FLAG_NONE 0
/* place entry at the head of the message */
#define SE_FLAG_PLACE_HEAD 1
/* place entry at the tail of the message */
#define SE_FLAG_PLACE_TAIL 2

#define SE_PLACEMENT_FLAG 3

/**
 * Entry in the send buffer.  Contains the size of the message, the
 * priority, when the message was passed to unicast, a callback to
 * fill in the actual message and a closure (argument to the
 * callback).
 **/
typedef struct {
  /** how long is this message part expected to be? */
  unsigned short len;
  /** flags */
  unsigned short flags;
  /** how important is this message part? */
  unsigned int pri;
  /** when did we intend to transmit? */
  cron_t transmissionTime;
  /** callback to call to create the message part */
  BuildMessageCallback callback;
  /** argument to callback, call FREENONNULL(closure) if we
      can not transmit this MessagePart. */
  void * closure;
} SendEntry;

/**
 * Type of the connection table.
 **/
typedef struct BufferEntry_ {
  /** Session for the connection */
  Session session;
  /** How much do we trust the host?  signed
      because that makes many operations that go 
      negative easier.  Of course, negative trust makes
      no sense. */
  int trust;
  /** the current session key */
  SESSIONKEY skey;
  /** at which time was the sessionkey created (by whichever party) */
  TIME_T created;
  /** is this host alive? timestamp of the time of the last-active point */
  cron_t isAlive; 
  /**  Status of the connection (STAT_XXX) **/
  unsigned int status;


  /** last sequence number received on this connection (highest) */
  unsigned int lastSequenceNumberReceived;
  /** bit map indicating which of the 32 sequence numbers before the last
      were received (good for accepting out-of-order packets and
      estimating reliability of the connection) */
  unsigned int lastPacketsBitmap;
  /** last sequence number transmitted */
  unsigned int lastSequenceNumberSend;

  /** number of entries in the send buffer */
  unsigned int sendBufferSize;
  
  /** buffer of entries waiting to be transmitted */
  SendEntry ** sendBuffer;

  /** time of the last send-attempt (to avoid
      solving knapsack's too often) */
  cron_t lastSendAttempt;

  /** a hash collision overflow chain **/
  struct BufferEntry_ * overflowChain;


  /* *********** outbound bandwidth limits ********** */

  /** byte-per-minute limit for this connection */
  unsigned int max_bpm;
  /** current bps (actually bytes per minute) for this connection
      (incremented every minute by max_bpm,
       bounded by max_bpm * secondsInactive/2;
       may get negative if we have VERY high priority
       content) */
  long long cur_bps;
  /** time of the last increment of cur_bps */
  cron_t last_bps_update;

  /* *********** inbound bandwidth accounting ******** */

  /* how much traffic (bytes) did we receive on this connection since
     the last update-round? */
  long long recently_received;

  /* what is the limit that we communicated last? (byte per minute) */
  unsigned int transmitted_limit;
  /* what is the limit that we are currently shooting for? (byte per minute) */
  unsigned int idealized_limit;

} BufferEntry;

/**
 * Type of a callback method on every buffer.
 * @param be the buffer entry
 * @param data context for callee
 **/
typedef void (*BufferEntryCallback)(BufferEntry * be,
				    void * data);

/* ***************** globals ********************** */

/**
 * The buffer containing all current connections.
 **/
static BufferEntry ** CONNECTION_buffer_;

/**
 * Size of the CONNECTION_buffer_
 **/
static unsigned int CONNECTION_MAX_HOSTS_;

/**
 * The DirContents array for scanning the hosts/ directory.
 **/
static unsigned int CONNECTION_currentActiveHosts;

/**
 * Experimental configuration: disable random padding of encrypted
 * messages.
 **/
static int disable_random_padding = NO;

/**
 * Send callbacks for making better use of noise padding...
 **/
static SendCallbackList * scl_nextHead;
static SendCallbackList * scl_nextTail;

/**
 * statistics handles
 **/
static int stat_number_of_connections;
static int stat_number_of_bytes_noise_send;
static int stat_number_of_bytes_send;
static int stat_number_of_bytes_received;
static int stat_sessionkeys_received;
static int stat_sessionkeys_verified;
static int stat_sessionkeys_transmitted;
static int stat_connections_shutdown;

/**
 * Lock for the connection module.
 **/
static Mutex lock;

/**
 * Where do we store trust information?
 **/
static char * trustDirectory;

/**
 * What is the available downstream bandwidth (in bytes
 * per minute)?
 **/
static long long max_bpm;

/* ******************** CODE ********************* */

/** 
 * This allocates and initializes a BufferEntry.
 * @return the initialized BufferEntry
 **/
static BufferEntry * initBufferEntry() {
  BufferEntry * be;

  be = (BufferEntry*) MALLOC(sizeof(BufferEntry));
  be->trust 
    = 0;
  be->isAlive 
    = 0;
  be->status 
    = STAT_DOWN;
  be->sendBuffer
    = NULL;
  be->sendBufferSize
    = 0;
  be->overflowChain
    = NULL;
  be->session.tsession
    = NULL;
  be->max_bpm
    = 65536 * 1024; /* 64 MB/minute is close enough to infinity */
  be->cur_bps
    = be->max_bpm;
  be->recently_received
    = 0;
  be->transmitted_limit
    = 65536 * 1024;
  be->lastSendAttempt
    = 0; /* never */
  cronTime(&be->last_bps_update); /* now */
  return be;
}

/**
 * Update cur_bps.  Call only when already synchronized.
 * @param be the connection for which to update cur_bps
 **/
void updateCurBPS(BufferEntry * be) {
  cron_t now;
  cron_t delta;

  cronTime(&now);
  delta = now - be->last_bps_update;
  be->cur_bps += (be->max_bpm * delta) / cronMINUTES;
  /* cap at reasonable max */
  if (be->cur_bps >
      be->max_bpm * cronSECONDS * SECONDS_INACTIVE_DROP / 2 / cronMINUTES)
    be->cur_bps 
      = be->max_bpm * cronSECONDS * SECONDS_INACTIVE_DROP / 2 / cronMINUTES;
  be->last_bps_update = now;    
}

/**
 * From time to time, do a recount on how many hosts are connected.
 **/
static void cronCountConnections() {
  unsigned int act;
  unsigned int i;
  BufferEntry * root;
  BufferEntry * tmp;

  act = 0;
  MUTEX_LOCK(&lock);
  for (i=0;i< CONNECTION_MAX_HOSTS_;i++) {
    root = CONNECTION_buffer_[i];

    tmp = root;
    while(NULL != tmp) {
      if (tmp->status == STAT_UP)
        act++;
      tmp = tmp->overflowChain;
    }
    
  }
  CONNECTION_currentActiveHosts = act;
  statSet(stat_number_of_connections,
	  act);
  MUTEX_UNLOCK(&lock);
}


/**
 * Write host-infromation to a file - flush the buffer entry!
 * Assumes synchronized access.
 **/
static void flushHostCredit(BufferEntry * be,
			    void * unused) {
  HexName fil;
  char * fn;

  if ((be->trust & 0x80000000) == 0)
    return; /* unchanged */
  be->trust = be->trust & 0x7FFFFFFF;
  hash2hex(&be->session.sender.hashPubKey,
	   &fil);
  fn = MALLOC(strlen((char*)trustDirectory)+sizeof(HexName)+1);
  buildFileName(trustDirectory,
		&fil,
		fn);
  writeFile(fn, 
	    &be->trust, 
	    sizeof(unsigned int), 
	    "644");
  FREE(fn);
}

/**
 * Compute the greatest common denominator (Euklid).
 *
 * @param a
 * @param b
 * @return gcd(a,b)
 **/
static int gcd(int a, int b) {
  while (a != 0) {
    int t = a;
    a = b % a;
    b = t;
  }
  return b;
}

/**
 * Solve the 0-1 knapsack problem.  Given "count" "entries" of
 * different "len" and "pri"ority and the amount of space "available",
 * compute the "solution", which is the set of entries to transport.
 * 
 * Solving this problem is NP complete in "count", but given that
 * available is small, the complexity is actually
 * "O(count*available)".
 *
 * @param entries the items to select from (priority and length are important)
 * @param count the number of entries in entries
 * @param available what is the maximum length available?
 * @param solution int[count] to store the solution as "YES" and "NO" values
 * @return the overall priority that was achieved 
 **/
static int solveKnapsack(SendEntry ** entries,
			 unsigned int count,
			 int available,
			 int * solution) {
  unsigned int i;
  int j;
  int max;
  long long * v;
  int * efflen;
  cron_t startTime;
  cron_t endTime;
#define VARR(i,j) v[(i)+(j)*(count+1)]

  ENTRY();
  cronTime(&startTime);
  max = 0;
  for (i=0;i<count;i++)
    max += entries[i]->len;
  if (max <= available) {
    /* short cut: take everything! */
    for (i=0;i<count;i++)
      solution[i] = YES;
    max = 0;
    for (i=0;i<count;i++)
      max += entries[i]->pri;
  }

  /* division of sizes & available by gcd(sizes,available) 
     to reduce cost to O(count*available/gcd) in terms of
     CPU and memory.  Since gcd is almost always at least 
     4, this is probably a good idea (TM)  :-) */
  efflen = MALLOC(sizeof(int)*count);
  max = available;
  for (i=0;i<count;i++)
    max = gcd(max, entries[i]->len);
  available = available / max;
  for (i=0;i<count;i++)
    efflen[i] = entries[i]->len / max;

  /* dynamic programming: 
     VARR(i,j) stores the maximum value of any subset
     of objects {1, ... i} that can fit into a knapsack
     of weight j. */
  v = MALLOC(sizeof(long long) * (count+1) * (available+1));
  memset(v, 
	 0, 
	 sizeof(long long) * (count+1) * (available+1));
  for (j=1;j<=available;j++)
    VARR(0,j) = -1;
  for (i=1;i<=count;i++) {
    for (j=0;j<=available;j++) {
      int take_val;
      int leave_val;

      take_val = -1;
      leave_val = VARR(i-1,j);
      if (j >= efflen[i-1]) {
	take_val = entries[i-1]->pri + VARR(i-1, j-efflen[i-1]);
	if (leave_val > take_val)
	  VARR(i,j) = leave_val;
	else
	  VARR(i,j) = take_val;
      } else
	VARR(i,j) = leave_val;
      /*
      printf("i: %d j: %d (of %d) efflen: %d take: %d "
             "leave %d e[i-1]->pri %d VAR(i-1,j-eff) %lld VAR(i,j) %lld\n",
	     i, 
	     j, 
	     available,
	     efflen[i-1],
	     take_val, 
	     leave_val,
	     entries[i-1]->pri,
	     VARR(i-1,j-efflen[i-1]),
	     VARR(i,j));
      */
    }
  }

  /* find slot with max value, prefer long messages! */
  max = 0;
  j = -1;
  for (i=0;(int)i<=available;i++) {
    if (VARR(count, i) >= max) {
      j = i;
      max = VARR(count, i);
    }
  }

  /* reconstruct selection */
  for (i=0;i<count;i++)
    solution[i] = NO;
  for (i=count;i>0;i--) {
    if (j >= efflen[i-1]) {
      if (VARR(i-1, j-efflen[i-1]) + entries[i-1]->pri == VARR(i,j)) {
	j -= efflen[i-1];
	solution[i-1] = YES; 
      }
    }
  }
  if (j != 0)
    errexit("FATAL: knapsack error\n");
  FREE(v);
  FREE(efflen);
  cronTime(&endTime);
  /* do not use Knapsack solver for more than 25% of the 
     CPU load that GNUnet is allowed to use. */
  MAX_SEND_FREQUENCY = (unsigned int) 
    ((endTime - startTime) *  CONNECTION_currentActiveHosts * 4 * (getCPULoad()/100.0));
    /* time it took here   *  average over all peers        * 25% * of what would be 100% CPU */
  return max;
}

/**
 * Send a buffer; assumes that access is already synchronized.  This
 * message solves the knapsack problem, assembles the message
 * (callback to build parts from knapsack, callbacks for padding,
 * random noise padding, crc, encryption) and finally hands the
 * message to the transport service.
 *
 * @param be connection of the buffer that is to be transmitted
 **/
static void sendBuffer(BufferEntry * be) {
  int crc;
  unsigned int i;
  unsigned int j;
  unsigned int p;
  SendCallbackList * pos;
  int priority;
  int * knapsackSolution;
  int * perm;
  char * plaintextMsg;
  void * encryptedMsg;
  cron_t expired;
  int targetSBLEN;
  SEQUENCE_Message * seqMsg;
  int headpos;
  int tailpos;

  ENTRY();
  /* fast ways out */
  if (be == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: sendBuffer called with invalid arguments (NULL).\n");
    return;
  }  
  if (be->status == STAT_DOWN)
    return; /* status is down, nothing to send! (should at least be wait-for-ping/pong 
	       or up) */
  if (be->sendBufferSize == 0) 
    return; /* nothing to send */  

  if (be->lastSendAttempt + MAX_SEND_FREQUENCY >
      cronTime(NULL))
    return; /* frequency too high, wait */
  /* solve knapsack problem, compute accumulated priority */
  knapsackSolution = MALLOC(sizeof(int) * be->sendBufferSize);
  priority = solveKnapsack(be->sendBuffer,
			   be->sendBufferSize,
			   be->session.mtu - sizeof(SEQUENCE_Message),
			   knapsackSolution);
  j = 0;
  for (i=0;i<be->sendBufferSize;i++)
    if (knapsackSolution[i] == YES)
      j++;
  if (j == 0) {
    LOG(LOG_ERROR,
	"ERROR: solveKnapsack selected %d out of %d messages (MTU: %d)\n",
	j,
	be->sendBufferSize,
	be->session.mtu - sizeof(SEQUENCE_Message));
    
    for (j=0;j<be->sendBufferSize;j++)
      LOG(LOG_ERROR,
	  "ERROR: %u: len %d, pri: %d\n",
	  j, 
	  be->sendBuffer[j]->len,
	  be->sendBuffer[j]->pri);
    FREE(knapsackSolution);
    return;
  }
    
  /* test if receiver has enough bandwidth available!  */
  updateCurBPS(be);
  if (be->cur_bps < be->session.mtu) {
    /* if we have a very high priority, we may
       want to ignore bandwidth availability (e.g. for HANGUP,
       which  has EXTREME_PRIORITY) */
    if (priority < EXTREME_PRIORITY) {
      FREE(knapsackSolution);
      /* LOG(LOG_INFO,
	 "INFO: bandwidth limits prevent sending\n");*/
      return; /* can not send, BPS available is too small */
    }
  }


  /* check if we (sender) have enough bandwidth available */
  if (SYSERR == outgoingCheck(priority)) {
    FREE(knapsackSolution);
    cronTime(&be->lastSendAttempt);
    /*LOG(LOG_INFO,
	"INFO: policy prevents sending message (prio %d)\n",
	priority);*/
    return; /* deferr further */
  }

  /* build message (start with sequence number) */
  plaintextMsg = MALLOC(be->session.mtu);
  seqMsg = (SEQUENCE_Message*) plaintextMsg;
  seqMsg->header.size 
    = htons(sizeof(SEQUENCE_Message));
  seqMsg->header.requestType 
    = htons(p2p_PROTO_SEQUENCE);
  seqMsg->sequenceNumber 
    = htonl(be->lastSequenceNumberSend);
  p = sizeof(SEQUENCE_Message);
  perm = permute(be->sendBufferSize);
  /* change permutation such that SE_FLAGS
     are obeyed */
  headpos = 0;
  tailpos = be->sendBufferSize-1;
  for (i=0;i<be->sendBufferSize;i++) 
    if (knapsackSolution[perm[i]] == YES) {
      switch (be->sendBuffer[perm[i]]->flags & SE_PLACEMENT_FLAG) {
      case SE_FLAG_NONE:
	break;
      case SE_FLAG_PLACE_HEAD:
	/* swap slot with whoever is head now */
	j = perm[headpos];
	perm[headpos++] = perm[i];
	perm[i] = j;
	break;
      case SE_FLAG_PLACE_TAIL:
	/* swap slot with whoever is tail now */
	j = perm[tailpos];
	perm[tailpos--] = perm[i];
	perm[i] = j;
      }
    }

  expired = cronTime(NULL) - SECONDS_INACTIVE_DROP; /* if it's more than one connection "lifetime" old, always kill it! */
  targetSBLEN = 0; /* how many entries in sendBuffer
		      afterwards? */
  for (i=0;i<be->sendBufferSize;i++) {
    SendEntry * entry = be->sendBuffer[perm[i]];
    if (knapsackSolution[perm[i]] == YES) {
      int ret;

      ret = entry->callback(&plaintextMsg[p],
			    entry->closure,
			    entry->len);
      if (ret == SYSERR) {	
	/* should not happen if everything went well,
	   add random padding instead */
	p2p_HEADER * part;

	part = (p2p_HEADER *) &plaintextMsg[p];
	part->size 
	  = htons(entry->len);
	part->requestType 
	  = htons(p2p_PROTO_NOISE);
	for (i=p+sizeof(p2p_HEADER);i<entry->len+p;i++)
	  plaintextMsg[p] = (char) rand();
	statChange(stat_number_of_bytes_noise_send,
		   entry->len);
      }
      p += entry->len;
      FREE(entry);
      be->sendBuffer[perm[i]] = NULL;
    } else if (entry->transmissionTime < expired) {
      FREE(entry->closure);
      FREE(entry);
      be->sendBuffer[perm[i]] = NULL;
    } else
      targetSBLEN++;
  }
  FREE(perm);
  FREE(knapsackSolution);

  /* cleanup/compact sendBuffer */
  j = 0;
  for (i=0;i<be->sendBufferSize;i++) 
    if (be->sendBuffer[i] != NULL)
      be->sendBuffer[j++] = be->sendBuffer[i];  
  GROW(be->sendBuffer,
       be->sendBufferSize,
       targetSBLEN);
    
  /* still room left? try callbacks! */
  pos = scl_nextHead;
  while (pos != NULL) {
    if (pos->minimumPadding + p <= be->session.mtu) 
      p += pos->callback(&be->session.sender,
			 &plaintextMsg[p],
			 be->session.mtu - p);
    pos = pos->next;
  }
  
  /* finally padd with noise */  
  if ( (p + sizeof(p2p_HEADER) <= be->session.mtu) &&
       (disable_random_padding == NO) ) {
    p2p_HEADER * part;
    unsigned short noiseLen = be->session.mtu - p;

    statChange(stat_number_of_bytes_noise_send,
	       noiseLen);
    part = (p2p_HEADER *) &plaintextMsg[p];
    part->size 
      = htons(noiseLen);
    part->requestType 
      = htons(p2p_PROTO_NOISE);
    for (i=p+sizeof(p2p_HEADER);
	 i < be->session.mtu;
	 i++)
      plaintextMsg[i] = (char) rand();
    p = be->session.mtu;
  }

  /* prepare for sending... */
  crc = crc32N(plaintextMsg,
	       p);
  encryptedMsg = MALLOC(p);
  if (p == (unsigned int) encryptBlock(plaintextMsg,
				       p,
				       &be->skey,
				       INITVALUE,
				       encryptedMsg)) {    
    statChange(stat_number_of_bytes_send,
	       p);
#if DEBUG_CONNECTION
    LOG(LOG_DEBUG,
	"DEBUG: calling transport layer to send %d bytes\n",
	p);
#endif
    transportSend(be->session.tsession,
		  encryptedMsg,
		  p,
		  YES, /* is encrypted */
		  crc);  
    be->lastSequenceNumberSend++;
  } else
    LOG(LOG_FAILURE,
	"FAILURE: Encryption of message failed!\n");
  FREE(encryptedMsg);
  FREE(plaintextMsg);
}

/**
 * Append a message to the current buffer. This method
 * assumes that the access to be is already synchronized.
 * 
 * @param be on which connection to transmit
 * @param se what to transmit (with meta-data)
 **/
static void appendToBuffer(BufferEntry * be,
			   SendEntry * se) {
  int i;
  int select;
  cron_t oldest;
  cron_t now;
#if DEBUG_CONNECTION
  HexName hex;
#endif  

  ENTRY();
  if (se == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: appendToBuffer got garbage. Ignored.\n");
    return;
  }
#if DEBUG_CONNECTION
  IFLOG(LOG_DEBUG,
	hash2hex(&be->session.sender.hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: adding message of size %d to buffer of host %s.\n",
      se->len,
      &hex);
#endif

  /* try selecting an "old" entry for replacement, then
     we don't have to grow! */
  select = -1;
  cronTime(&now);
  oldest = now - SECONDS_INACTIVE_DROP / 2; /* never discard anything that is younger
					       than 1/2 of the inactive period */
  if (be->sendBufferSize > TARGET_SBUF_SIZE)
    for (i=be->sendBufferSize-1;i>=0;i--) 
      if (be->sendBuffer[i]->transmissionTime < oldest) {
	select = i;
	oldest = be->sendBuffer[i]->transmissionTime;
      }
  if (select != -1) {
    FREENONNULL(be->sendBuffer[select]->closure);
    FREE(be->sendBuffer[select]);
    be->sendBuffer[select] = se;
  } else {
    /* no "old" entry present, grow! */
    GROW(be->sendBuffer,
	 be->sendBufferSize,
	 be->sendBufferSize+1);
    be->sendBuffer[be->sendBufferSize-1] = se;
  }

  /* call sendBuffer if any of the entries has
     expired already */
  oldest = now + 1; /* infty */
  for (i=be->sendBufferSize-1;i>=0;i--) 
    if (be->sendBuffer[i]->transmissionTime < oldest) 
      oldest = be->sendBuffer[i]->transmissionTime;
      
 /*  LOG(LOG_DEBUG,
      "DEBUG: deciding on sendBuffer call: %lld <= %lld? %u\n",
      oldest, now,
      be->sendBufferSize);*/
  if (oldest <= now) {
/*    LOG(LOG_DEBUG,
	"DEBUG: calling sendBuffer!\n");*/
    sendBuffer(be);
  }
}

/**
 * Look for a host in the table. If the entry is there at the time of
 * checking, returns the entry.
 *
 * @param hostId the ID of the peer for which the connection is returned
 * @return the connection of the host in the table, NULL if not connected
 **/
static BufferEntry * lookForHost(const HostIdentity * hostId) {
  BufferEntry * root;

  root = CONNECTION_buffer_[computeIndex(hostId)];
  while (root != NULL) {
    if (equalsHashCode160(&hostId->hashPubKey,
			  &root->session.sender.hashPubKey)) 
      return root;
    root = root->overflowChain;
  }
  return NULL;
}


/**
 * Read host-information from a file.  The connection lock
 * must be held.
 * 
 * @param be connection of the peer for which the trust is to be read
 **/
static void initHostTrust(BufferEntry * be) {
  HexName fil;
  char * fn;

  hash2hex(&be->session.sender.hashPubKey,
	   &fil);
  fn = MALLOC(strlen((char*)trustDirectory)+sizeof(HexName)+1);
  buildFileName(trustDirectory, &fil, fn); 
  if (sizeof(unsigned int) !=
      readFile(fn, 
	       sizeof(unsigned int), 
	       &be->trust))
    be->trust = 0;
  FREE(fn);
}

/**
 * Force adding of a host to the buffer. If the node is already in the
 * table, the table entry is returned.  
 *
 * @param hostId for which peer should we get/create a connection
 * @param force if YES, drop another host from the table if the slot
 * is already in use. If NO, return NULL if the slot is busy.  
 * @return the table entry for the host (no keyexchange performed so far)
 **/
static BufferEntry * addHost(const HostIdentity * hostId,
			     int force) {
  BufferEntry * root;
  BufferEntry * prev;
  HexName hex;

  IFLOG(LOG_INFO,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_INFO, 
      "INFO: Adding host %s to the connection table.\n",
      &hex);
#endif

  ENTRY();
  root = lookForHost(hostId);
  if (root != NULL)
    return root;

  root = CONNECTION_buffer_[computeIndex(hostId)];
  prev = NULL;
  while (NULL != root) {
    /* settle for entry in the linked list that is down */
    if ( (root->status == STAT_DOWN) ||
	 (equalsHashCode160(&hostId->hashPubKey,
			    &root->session.sender.hashPubKey)) ) 
      break;
    prev = root;
    root = root->overflowChain;
  }
  if (root == NULL) {
    root = initBufferEntry();
    if (prev == NULL)
      CONNECTION_buffer_[computeIndex(hostId)] = root;
    else
      prev->overflowChain = root;
  }
  memcpy(&root->session.sender,
	 hostId,
	 sizeof(HostIdentity));	 
  initHostTrust(root);
  return root;
}

/**
 * Perform an operation for all connected hosts.  The BufferEntry
 * structure is passed to the method.  No synchronization or other
 * checks are performed.
 *
 * @param method the method to invoke (NULL for couting only)
 * @param arg the second argument to the method
 * @return the number of connected hosts
 **/ 
static int forAllConnectedHosts(BufferEntryCallback method,
				void * arg) {
  unsigned int i;
  int count = 0;
  BufferEntry * be;
  
  ENTRY();
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = CONNECTION_buffer_[i];
    while (be != NULL) {
      if (be->status == STAT_UP) {
        if (method != NULL)
  	  method(be, arg);
        count++;
      }
      be = be->overflowChain;
    }
  }
  return count;
}

/**
 * Little helper function for forEachConnectedNode.
 * 
 * @param be the connection 
 * @param arg closure of type fENHWrap giving the function
 *        to call
 **/
static void fENHCallback(BufferEntry * be,
			 void * arg) {
  fENHWrap * wrap;

  wrap = (fENHWrap*) arg;
  if (wrap->method != NULL)
    wrap->method(&be->session.sender,
		 wrap->arg);
}


/**
 * Here in this scanning for applicable hosts, we also want to take
 * the protocols into account and prefer "cheap" protocols,
 * i.e. protocols with a low overhead.
 *
 * @param id which peer are we currently looking at
 * @param proto what transport protocol are we looking at
 * @param im updated structure used to select the peer
 **/
static void scanHelperCount(const HostIdentity * id,
			    const unsigned short proto,	
			    IndexMatch * im) {
  if (hostIdentityEquals(&myIdentity, id)) 
    return;
  if (computeIndex(id) != im->index)
    return;
  if (YES == isTransportAvailable(proto)) {
    im->matchCount++;  
    im->costSelector += transportGetCost(proto);
  }
#if DEBUG_CONNECTION
  else
    LOG(LOG_DEBUG,
	"DEBUG: transport %d is not available\n",
	proto);
#endif
}

/**
 * Select the peer and transport that was selected based on transport
 * cost.
 * 
 * @param id the current peer
 * @param proto the protocol of the current peer
 * @param im structure responsible for the selection process
 **/
static void scanHelperSelect(const HostIdentity * id,
			     const unsigned short proto,
			     IndexMatch * im) {
  if (hostIdentityEquals(&myIdentity, id)) 
    return;
  if (computeIndex(id) != im->index)
    return;
  if (YES == isTransportAvailable(proto)) {
    im->costSelector -= transportGetCost(proto);
    if ( (im->matchCount == 0) ||
	 (im->costSelector < 0) ) {
      memcpy(&im->match,
	     id,
	     sizeof(HostIdentity));
    }
    im->matchCount--;
  }
#if DEBUG_CONNECTION
  else
    LOG(LOG_DEBUG,
	"DEBUG: transport %d is not available\n",
	proto);
#endif
}

/**
 * Force creation of a new Session key for the given host. 
 *
 * @param hostId the identity of the other host 
 * @param sk the SESSIONKEY to use
 * @param created the timestamp to use
 * @param ret the address where to write the signed
 *        session key, first unsigned short_SIZE byte give length
 * @return OK on success, SYSERR on failure
 **/
static int makeSessionKeySigned(HostIdentity * hostId,
				SESSIONKEY * sk,
				TIME_T created,
				SKEY_Message * ret) {
  HashCode160 keyHash;
  HostIdentity myId;
  HELO_Message * foreignHelo;
#if EXTRA_CHECKS
  HexName hostName;
#endif

  ENTRY();
  if ( (ret == NULL) ||
       (sk == NULL) )
    errexit("FATAL: makeSessionKeySigned called with"
	    " ret or sk equal to NULL"
	    " (interface violation)\n");
  /* create and encrypt sessionkey */
  if (SYSERR == identity2Helo(hostId, 
			      ANY_PROTOCOL_NUMBER,
			      YES,
			      &foreignHelo)) {
    LOG(LOG_WARNING, 
	"WARNING: makeSesionKeySigned: cannot encrypt"
	" sessionkey, other host not known!\n");  
    return SYSERR; /* other host not known */  
  }
  if (foreignHelo == NULL)
    errexit("FATAL: identity2Helo violated interface, "
	    "returned OK but did not set helo ptr\n");
  if (SYSERR == encryptHostkey(sk,
			       sizeof(SESSIONKEY),
			       &foreignHelo->publicKey,
			       &ret->body.key)) {
    LOG(LOG_FAILURE, 
	"FAILURE: makeSesionKeySigned: encrypt failed!\n");
    FREE(foreignHelo);
    return SYSERR; /* encrypt failed */
  }
  FREE(foreignHelo);
  /* compute hash and sign hash */
  ret->body.creationTime = htonl(created);
  hash(&ret->body, 
       sizeof(RSAEncryptedData) + sizeof(TIME_T),
       &keyHash);
  if (SYSERR == signData(&keyHash, 
			 (unsigned short)sizeof(HashCode160),
			 &ret->body.signature)) {
    LOG(LOG_FAILURE, 
	"FAILURE: signData failed. Trying to continue.\n");
  }

  /* complete header */
  ret->header.size = htons(sizeof(SKEY_Message));
  ret->header.requestType = htons(p2p_PROTO_SKEY);
  /* verify signature/SKS */
  getHostIdentity(getPublicHostkey(),
		  &myId); 
#if EXTRA_CHECKS
  hash2hex(&hostId->hashPubKey, &hostName);
  if (SYSERR == verifySKS(&myId, ret)) {
    errexit("FATAL: makeSessionKeySigned: generated sessionkey for host"
	    " %s failed verification! exiting!\n",
	    &hostName);
  }
#endif
  return OK;
} 

/**
 * Perform a session key exchange for entry be.  First sends a HELO
 * and then the new SKEY (in two plaintext packets). When called, the
 * semaphore of at the given index must already be down
 *
 * @param be connection on which the key exchange is performed
 **/
static void exchangeKey(BufferEntry * be) {
  HexName hex;
  HELO_Message * helo;
  HELO_Message * targetHelo;
  SKEY_Message skey;
  unsigned short targetTransport;
  char * sendBuffer;

  ENTRY();
  if (be == NULL)
    errexit("FATAL: exchangeKey called with be==NULL\n");
  IFLOG(LOG_DEBUG,
	hash2hex(&be->session.sender.hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: Beginning of exchange key with %s\n",
	&hex);
#endif
  if (be->status != STAT_DOWN) 
    LOG(LOG_WARNING, 
	"WARNING: exchangeKey: status must"
	" be down for session key exchange!\n");

  makeSessionkey(&be->skey);
  TIME(&be->created); /* set creation time for session key! */
  if (SYSERR == makeSessionKeySigned(&be->session.sender,
				     &be->skey,
				     be->created,
				     &skey))
    return;
  be->isAlive = 0; /* wait a bit */
  be->status = STAT_WAITING_FOR_PING;
  be->lastSequenceNumberReceived = 0;
  be->lastPacketsBitmap = (unsigned int) -1;
  /* send tmp to target */
  if (SYSERR == identity2Helo(&be->session.sender,
			      ANY_PROTOCOL_NUMBER,
			      YES,
			      &targetHelo)) {
    return;
  }
  targetTransport = ntohs(targetHelo->protocol);
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG,
      "DEBUG: identity2Helo returned HELO with protocol %d\n",
      targetTransport);
#endif
  if (SYSERR == transportCreateHELO(targetTransport,
				    &helo)) {
    be->status = STAT_DOWN;
    FREE(targetHelo);
    return;
  }
  if (SYSERR == transportConnect(targetHelo, /* callee frees except on SYSERR */
				 &be->session.tsession)) {
    be->status = STAT_DOWN;
    be->session.tsession = NULL;
    FREE(targetHelo);
    FREE(helo);
    return;
  } 
  targetHelo = NULL; /* ensure that we do not use the helo, 
			transportConnect is now the owner! */
  be->session.mtu = transportGetMTU(be->session.tsession->ttype);
  if (be->sendBuffer != NULL)
    LOG(LOG_FAILURE,
	"FAILURE: expected be->sendBuffer to be NULL in file %s:%d\n",
	__FILE__, __LINE__);
  be->lastSequenceNumberSend = 1;
  be->session.isEncrypted = NO;

  sendBuffer = MALLOC(HELO_Message_size(helo)+
		      sizeof(SKEY_Message));
  memcpy(sendBuffer,
	 helo,
	 HELO_Message_size(helo));
  memcpy(&sendBuffer[HELO_Message_size(helo)],
	 &skey,
	 sizeof(SKEY_Message));
  
  transportSend(be->session.tsession,
		sendBuffer,
		HELO_Message_size(helo)+sizeof(SKEY_Message),
		NO, /* not encrypted */
		crc32N(sendBuffer, HELO_Message_size(helo)+sizeof(SKEY_Message)));
  FREE(sendBuffer);
  FREE(helo);
  be->session.isEncrypted = YES;
  statChange(stat_sessionkeys_transmitted, 1);
}

/**
 * Look in the list for known hosts; pick a random host of minimal
 * transport cost for the hosttable at index index. When called, the
 * mutex of at the given index must not be hold.
 *
 * @param index for which entry in the connection table are we looking for peers?
 **/
static void scanForHosts(unsigned int index) {
  BufferEntry * be;
  IndexMatch indexMatch;
  HexName hn;
  cron_t now;

#if DEBUG_CONNECTION
  LOG(LOG_CRON, 
      "CRON: Scanning for hosts (%d)\n",
      index);
#endif
  cronTime(&now);
  indexMatch.index = index;
  indexMatch.matchCount = 0;
  indexMatch.costSelector = 0;
  forEachHost((HostIterator)&scanHelperCount, 
	      now,
	      &indexMatch);
  if (indexMatch.matchCount == 0) 
    return;  
  LOG(LOG_CRON, 
      "CRON: Scanning for hosts (%d) found %d matching node identities.\n",
      index,
      indexMatch.matchCount);    
  if (indexMatch.costSelector > 0)
    indexMatch.costSelector
      = randomi(indexMatch.costSelector/4)*4;
  memcpy(&indexMatch.match,
	 &myIdentity,
	 sizeof(HostIdentity));  
  forEachHost((HostIterator)&scanHelperSelect,
	      now,
	      &indexMatch);
  if (hostIdentityEquals(&myIdentity, 
			 &indexMatch.match)) {
    LOG(LOG_ERROR, 
	"ERROR: scanHelperSelect did not find selected match!?\n");
    return;
  }
  if (computeIndex(&indexMatch.match) != index) {
    LOG(LOG_ERROR, 
	"ERROR: index of host to add is wrong!\n");
    return;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&indexMatch.match.hashPubKey,
		 &hn));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: attempting to connect to %s using slot %d.\n",
      &hn, 
      index);
#endif
  be = addHost(&indexMatch.match, 
	       NO); /* NO or YES should not matter here since we know the slot is empty */
  if (be != NULL) {
    if (be->status == STAT_DOWN) {
      blacklistHost(&be->session.sender,
		    CONNECTION_currentActiveHosts); 
      /* we're trying now, don't try again too soon */
      exchangeKey(be);
    }
  }
}

/**
 * Copy the pre-build message part of lenth "len" in closure to the
 * buffer buf. Frees the closure.
 *
 * @param buf the target location where the message is assembled
 * @param closure the pre-build message
 * @param len the length of the pre-build message
 * @return OK (always successful)
 **/
static int copyCallback(void * buf,
			void * closure,
			unsigned short len) {
  memcpy(buf, closure, len);
  FREE(closure);
  return OK; 
}

/**
 * Check if the buffer is up (we got a PONG), if not, repeat the PING 
 **/
static void checkAndPing(BufferEntry * be) {
  HostIdentity * data;
  PINGPONG_Message pmsg;

  ENTRY();
  data = MALLOC(sizeof(HostIdentity));
  memcpy(data,
	 &be->session.sender,
	 sizeof(HostIdentity));
  if (OK == pingAction(&be->session.sender,
		       (CronJob) &notifyPONG,
		       data,
		       &pmsg)) {
    SendEntry * se;
    
    se = MALLOC(sizeof(SendEntry));
    se->flags = SE_FLAG_NONE;
    se->len = sizeof(PINGPONG_Message);
    se->pri = getConnectPriority();
    se->transmissionTime = cronTime(NULL); /* now */
    se->callback = &copyCallback;
    se->closure = MALLOC(sizeof(PINGPONG_Message));
    memcpy(se->closure,
	   &pmsg,
	   sizeof(PINGPONG_Message));
    appendToBuffer(be, se);
  } else {
    FREE(data);
    LOG(LOG_WARNING,
	"WARNING: could not send checking ping, ping buffer full\n");
  }
}

/**
 * Shutdown the connection.  Send a HANGUP message to the other side
 * and mark the sessionkey as dead.
 *
 * @param be the connection to shutdown
 **/
static void shutdownConnection(BufferEntry * be) {
  HANGUP_Message hangup;
  unsigned int i;

  ENTRY();
  if (be->status == STAT_DOWN)
    return; /* nothing to do */
  if (be->status == STAT_UP) {
    SendEntry * se;

    hangup.header.requestType = htons(p2p_PROTO_HANGUP);
    hangup.header.size = htons(sizeof(HANGUP_Message));
    getHostIdentity(getPublicHostkey(),
		    &hangup.sender);    
    se = MALLOC(sizeof(SendEntry));
    se->len = sizeof(HANGUP_Message);
    se->flags = SE_FLAG_PLACE_TAIL;
    se->pri = EXTREME_PRIORITY;
    se->transmissionTime = cronTime(NULL); /* now */
    se->callback = &copyCallback;
    se->closure = MALLOC(sizeof(HANGUP_Message));
    memcpy(se->closure,
	   &hangup,
	   sizeof(HANGUP_Message));
    appendToBuffer(be, se);
  }
  be->created = 0;
  be->status = STAT_DOWN;
  if (be->session.tsession != NULL) {
    transportDisconnect(be->session.tsession);
    be->session.tsession = NULL;
  }    
  statChange(stat_connections_shutdown, 1);
  for (i=0;i<be->sendBufferSize;i++) {
    FREENONNULL(be->sendBuffer[i]->closure);
    FREE(be->sendBuffer[i]);
  }
  GROW(be->sendBuffer,
       be->sendBufferSize,
       0);
}

/**
 * Transmit an update to the bandwidth limit to the other peer.
 *
 * @param be the connection to transmit the limit over
 **/
static void transmitConnectionLimit(BufferEntry * be) {
  SendEntry * entry;
  CAPABILITY_Message * cap;
  
  ENTRY();
  entry = MALLOC(sizeof(SendEntry));
  entry->len = sizeof(CAPABILITY_Message);
  entry->flags = SE_FLAG_NONE;
  entry->pri = ADMIN_PRIORITY;
  entry->transmissionTime = cronTime(NULL);
  entry->callback = &copyCallback;
  entry->closure = MALLOC(sizeof(CAPABILITY_Message));
  cap = (CAPABILITY_Message*) entry->closure;
  cap->header.size = htons(sizeof(CAPABILITY_Message));
  cap->header.requestType = htons(p2p_PROTO_CAPABILITY);
  cap->cap.capabilityType = htonl(CAP_BANDWIDTH_RECV);
  cap->cap.value = htonl(be->idealized_limit);
  appendToBuffer(be, 
		 entry);   
  be->transmitted_limit = be->idealized_limit;
}

/**
 * Call this method periodically to decrease liveness of hosts.
 *
 * @param unused not used, just to make signature type nicely
 **/
static void cronDecreaseLiveness(void * unused) {  
  static unsigned int lastLivenessHost = 0;
  static cron_t lastRoundStart = 0;
  static cron_t timeDifference;
  static unsigned long long lastRoundRecv = 0;
  static int bandwidth_cap_change = 0;
  static int activePeerCount = 0;
  long long bpm;
  BufferEntry * root;
  BufferEntry * prev;
  BufferEntry * tmp;
  cron_t now;
  cron_t oldest;
  int i;  

  cronTime(&now);    
  if (lastLivenessHost == 0) {
    /* start a new round of bandwidth assignments */
    timeDifference = now - lastRoundStart;
    bpm = lastRoundRecv * cronMINUTES / timeDifference;
    if (activePeerCount > 0) {
      /* compute if we want more or less traffic overall */
      bandwidth_cap_change = (bpm - max_bpm) / activePeerCount; 
    } else {
      /* no active peers? well,
	 then we clearly want to discover some :-) */
      static int delay = -2; /* wait 2 rounds of inactivity at
				the beginning, just in case */
      bandwidth_cap_change = 0;
      /* at most try every 8 minutes */
      if (delay % 16 == 0)
	downloadHostlist();
      delay++;
    }

    lastRoundRecv = 0;
    activePeerCount = 0;
    cronTime(&lastRoundStart);
  }
  /* Find the correct host */
  MUTEX_LOCK(&lock);
  root = CONNECTION_buffer_[lastLivenessHost];
  prev = NULL;
  while (NULL != root) {
    /* is this the host? */
    switch (root->status) {
    case STAT_DOWN:
      if (prev == NULL)
	CONNECTION_buffer_[lastLivenessHost] = root->overflowChain;
      else
	prev->overflowChain = root->overflowChain;
      tmp = root;
      root = root->overflowChain;
      FREE(tmp);
      continue;
    case STAT_UP:
      if (bandwidth_cap_change != 0) {
	long long delta;
	unsigned int div;

	root->idealized_limit += bandwidth_cap_change;
	/* don't increase the max too high if the peer
	   does not send that much */
	if (root->idealized_limit >
	    2 * root->recently_received * cronMINUTES / timeDifference)
	  root->idealized_limit = 2 * root->recently_received * cronMINUTES / timeDifference;
	
	/* cap idealized_limit at the lowest *reasonable*
	   value (to keep connection alive!): allow at
	   least 16 MTU sized messages in every SECONDS_INACTIVE_DROP
	   interval. */
	div = (SECONDS_INACTIVE_DROP/60) * TARGET_MSG_SID; 
	/* minutes inactive * NUMBER_OF_MESSAGES */
	if (div <= 0)
	  div = 1; /* avoid FPE */
	if (root->idealized_limit < root->session.mtu / div) {
	  root->idealized_limit = root->session.mtu / div;
	  if (root->idealized_limit <= 0)
	    root->idealized_limit = 1; /* enforce sane minimum */
	}

	/* send update message on CAPABILITIES if they
	   change significantly */
	delta = root->idealized_limit - root->transmitted_limit;
	if (delta < 0)
	  delta = -delta;
	if ( ( (delta * 100) / root->transmitted_limit) > 10)				  
	  /* limits changed by more than 10%,
	     send updated capability message! */
	  transmitConnectionLimit(root);	             
      }
      lastRoundRecv += root->recently_received;
      root->recently_received = 0;
      activePeerCount++;
      if (now - root->isAlive > SECONDS_INACTIVE_DROP * cronSECONDS) {
	/* switch state form UP to DOWN: too much inactivity */
	LOG(LOG_DEBUG,
	    "DEBUG: closing connection: too much inactivity\n");
	shutdownConnection(root);
	whitelistHost(&root->session.sender); /* the host may still be worth trying again soon */
	CONNECTION_currentActiveHosts--;
	statChange(stat_number_of_connections,
		   -1);
	break;
      } 
      if ( (CONNECTION_currentActiveHosts < CONNECTION_MAX_HOSTS_/2) &&
	   (now - root->isAlive > SECONDS_PINGATTEMPT * cronSECONDS) ){
	/* if we have less than 50% of the number of connections
	   that we would like to have, try ping-ing the other side
	   to keep the connection open instead of hanging up */
	PINGPONG_Message pmsg;
	HostIdentity * hi = MALLOC(sizeof(HostIdentity));
	memcpy(hi, 
	       &root->session.sender, 
	       sizeof(HostIdentity));
	if (OK == pingAction(&root->session.sender,
			     (CronJob)&notifyPING,
			     hi,
			     &pmsg)) {
	  SendEntry * entry;

	  entry = MALLOC(sizeof(SendEntry));
	  entry->len = sizeof(PINGPONG_Message);
	  entry->flags = SE_FLAG_NONE;
	  entry->pri = getConnectPriority(); 
	  entry->transmissionTime = now + (now - root->isAlive)/2;
	  entry->callback = &copyCallback;
	  entry->closure = MALLOC(sizeof(PINGPONG_Message));
	  memcpy(entry->closure,
		 &pmsg,
		 sizeof(PINGPONG_Message));	  
	  appendToBuffer(root, 
			 entry);
	} else
	  FREE(hi);
      }
      break;
    case STAT_WAITING_FOR_PING:
      if (now - root->isAlive > SECONDS_NOPINGPONG_DROP * cronSECONDS) {
        HexName hex;
	IFLOG(LOG_DEBUG, hash2hex(&root->session.sender.hashPubKey, &hex));
	LOG(LOG_DEBUG,
	    "DEBUG: closing connection to %s: SKEY not answered by PING\n",
	    &hex);
	shutdownConnection(root);      
      }
      break;
    case STAT_WAITING_FOR_PONG:
      if (now - root->isAlive > SECONDS_NOPINGPONG_DROP * cronSECONDS) {
	HexName hex;
	IFLOG(LOG_DEBUG, hash2hex(&root->session.sender.hashPubKey, &hex));
	LOG(LOG_DEBUG,
	    "DEBUG: closing connection to %s: PING not answered by PONG\n",
	    &hex);
	shutdownConnection(root);      
      } else
	checkAndPing(root);
      break;
    default:
      LOG(LOG_WARNING,
	  "WARNING: unknown type of connection status: %d\n",
	  root->status);
      break; /* do nothing */
    } /* end of switch */ 
  
    /* call sendBuffer if any of the entries has
       expired already */
    oldest = now + 1; /* infty */
    for (i=root->sendBufferSize-1;i>=0;i--) 
      if (root->sendBuffer[i]->transmissionTime < oldest) 
	oldest = root->sendBuffer[i]->transmissionTime;
    /*    LOG(LOG_DEBUG,
	"DEBUG: deciding on sendBuffer call: %lld <= %lld? at %s:%d - %u\n",
	oldest, now,
	__FILE__, __LINE__, root->sendBufferSize);*/
    if (oldest <= now) {
      /* LOG(LOG_DEBUG,
	 "DEBUG: calling sendBuffer!\n");*/
      sendBuffer(root);      
    }

    prev = root;
    root = root->overflowChain;
  } /* end of while */
  if (CONNECTION_buffer_[lastLivenessHost] == NULL)
    scanForHosts(lastLivenessHost);

  lastLivenessHost++;
  if (lastLivenessHost >= CONNECTION_MAX_HOSTS_)
    lastLivenessHost = 0;
  MUTEX_UNLOCK(&lock);
}

/**
 * Check the sequence number.  Updates the sequence number as a
 * side-effect.
 *
 * @param sender from which peer did we receive the SEQ message
 * @param msg the sequence message
 * @returns OK if ok, SYSERR if not.
 **/
static int checkSequenceNumber(HostIdentity * sender,
			       p2p_HEADER * msg) {
  SEQUENCE_Message * smsg;
  BufferEntry * be;
  int res;
  unsigned int sequenceNumber;
  
  ENTRY();
  if (ntohs(msg->size) != sizeof(SEQUENCE_Message)) {
    LOG(LOG_WARNING,
	"WARNING: sequence message received has wrong size %d\n",
	ntohs(msg->size));
    return SYSERR;
  }
  smsg = (SEQUENCE_Message*) msg;
  sequenceNumber = ntohl(smsg->sequenceNumber);
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: sequence message received for node that is not connected!?\n");
    MUTEX_UNLOCK(&lock);
    return SYSERR; /* host not found */
  }
  res = OK;
  if (be->lastSequenceNumberReceived >= sequenceNumber) {
    unsigned int rotbit = 1;
    if ( (be->lastSequenceNumberReceived - sequenceNumber <= 32) && 
	 (be->lastSequenceNumberReceived != sequenceNumber) ) {
      rotbit = rotbit << (be->lastSequenceNumberReceived - sequenceNumber - 1);
      if ( (be->lastPacketsBitmap & rotbit) == 0) {
	res = OK;
	be->lastPacketsBitmap |= rotbit;
      } else
	res = SYSERR;
    } else
      res = SYSERR;
    if (res == SYSERR) {
      LOG(LOG_WARNING,
	  "WARNING: Invalid sequence number"
	  " %u < %u, dropping rest of packet\n",
	  sequenceNumber, 
	  be->lastSequenceNumberReceived);
    }    
  } else {    
    be->lastPacketsBitmap = 
      be->lastPacketsBitmap << (sequenceNumber - be->lastSequenceNumberReceived);
    be->lastSequenceNumberReceived = sequenceNumber;
  }
  MUTEX_UNLOCK(&lock);
  if (res == SYSERR)
    LOG(LOG_INFO,
	"INFO: message received has old sequence number. Dropped.\n");
  return res;
}

/**
 * We received a request from a client to provide the number
 * of directly connected peers.  Sends the response.
 * 
 * @param client the socket connecting to the client
 * @param msg the request from the client
 * @returns OK if ok, SYSERR if not.
 **/
static int processGetConnectionCountRequest(GNUNET_TCP_SOCKET * client,
					    CS_HEADER * msg) {
  if (ntohs(msg->size) != sizeof(CS_HEADER)) {
    LOG(LOG_WARNING,
	"WARNING: request received from TCP has wrong size %d\n",
	ntohs(msg->size));
    return SYSERR;
  }
  return sendTCPResult(client,
		       CONNECTION_currentActiveHosts);
}


/**
 * Handler for processing P2P HANGUP message.  Terminates
 * a connection (if HANGUP message is valid).
 *
 * @param sender the peer sending the HANGUP message
 * @param msg the HANGUP message
 * @return OK on success, SYSERR on error
 **/
static int handleHANGUP(HostIdentity * sender,
			p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hex;

  ENTRY();
  if (ntohs(msg->size) != sizeof(HANGUP_Message))
    return SYSERR;
  if (!hostIdentityEquals(sender,
			  &((HANGUP_Message*)msg)->sender))
    return SYSERR;
  IFLOG(LOG_INFO,
	hash2hex(&sender->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_INFO,
      "INFO: received HANGUP from %s\n",
      &hex);
#endif
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) {
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }
  LOG(LOG_DEBUG,
      "DEBUG: closing connection, received HANGUP\n");
  shutdownConnection(be);
  MUTEX_UNLOCK(&lock);
  return OK;
}

/**
 * Handler for processing CAPABILITY.  Resets the bandwidth
 * limit of a connection (or other capabilities that we may
 * define in the future).
 *
 * @param sender peer that send the CAP message
 * @param msg the CAP message
 * @return OK on success, SYSERR on error
 **/
static int handleCAPABILITY(HostIdentity * sender,
			    p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hex;
  CAPABILITY_Message * cap;

  ENTRY();
  if (ntohs(msg->size) != sizeof(CAPABILITY_Message))
    return SYSERR;
  cap =  (CAPABILITY_Message*)msg;
  IFLOG(LOG_INFO,
	hash2hex(&sender->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_INFO,
      "INFO: received CAPABILITY from %s\n",
      &hex);
#endif
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) {
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }
  switch (ntohl(cap->cap.capabilityType)) {
  case CAP_BANDWIDTH_RECV:
    be->max_bpm = ntohl(cap->cap.value);
    if (be->cur_bps >= be->max_bpm) {
      be->cur_bps = be->max_bpm;
      cronTime(&be->last_bps_update);
    }
    break;
  default:
    LOG(LOG_INFO,
	"INFO: capability type unknown: %d, ignored\n",
	ntohl(cap->cap.capabilityType));
    break;
  }
  MUTEX_UNLOCK(&lock);  
  return OK;
}


/**
 * Check if the received session key is properly signed.
 *
 * @param hostId the sender of the key
 * @param sks the session key message
 * @return SYSERR if invalid, OK if valid
 **/
static int verifySKS(const HostIdentity * hostId,
		     SKEY_Message * sks) {
  HashCode160 keyHash;
  HexName hostName;
  HELO_Message * helo;
  
  ENTRY();
  if ( (sks == NULL) ||
       (hostId == NULL)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&hostId->hashPubKey, 
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: VerifySKS: invalid sessionkey suggestion for host %s\n",
	&hostName);
    return SYSERR;
  }
  
  if (SYSERR == identity2Helo(hostId, 
			      ANY_PROTOCOL_NUMBER,
			      YES,
			      &helo)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&hostId->hashPubKey, 
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: verifySKS: host %s for sessionkey exchange not known\n",
	&hostName);
    return SYSERR;
  }
  /* verify signature */
  hash(&sks->body, 
       sizeof(RSAEncryptedData) + sizeof(TIME_T),
       &keyHash);
  if (!verifySig(&keyHash,
		 sizeof(HashCode160),
		 &sks->body.signature, 
		 &helo->publicKey)) {
    LOG(LOG_WARNING, 
	"WARNING: verifySKS: received SessionKey with bad signature!\n");
    FREE(helo);
    return SYSERR; /*reject!*/
  }
  FREE(helo);
  statChange(stat_sessionkeys_verified, 1);
  return OK; /* ok */
}

/**
 * Call once in a while to synchronize trust values with the disk.
 **/
static void cronFlushTrustBuffer(void * unused) {
#if DEBUG_CONNECTION
  LOG(LOG_CRON,
      "CRON: enter cronFlushTrustBuffer\n");
#endif
  MUTEX_LOCK(&lock);
  forAllConnectedHosts(&flushHostCredit, unused);
  MUTEX_UNLOCK(&lock);
#if DEBUG_CONNECTION
  LOG(LOG_CRON, 
      "CRON: exit cronFlushTrustBuffer\n");
#endif
}

/**
 * Connect to another peer.
 *
 * @param hostId the peer to connect with
 * @return the connection handle
 **/
static BufferEntry * connectTo(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  ENTRY();
  if (hostIdentityEquals(&myIdentity, 
			 hostId)) {
    LOG(LOG_ERROR, 
	"ERROR: attempt to connect to myself! (dropped)\n");
    BREAK();
    return NULL;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
  be = lookForHost(hostId);
  if ( (be == NULL) || 
       (be->status == STAT_DOWN) ) {
    be = addHost(hostId, YES);   /* we *really* want to talk to this guy */
    if (be->status == STAT_DOWN)
      exchangeKey(be); /* note that exchangeKey
			  can fail and leave the buffer
			  in "down" state, e.g. if we don't
			  know any key for the other peer */
  }
  return be;
}

/**
 * How important is it at the moment to establish more connections?
 *
 * @return a measure of the importance to establish connections
 **/
int getConnectPriority() {
  if (CONNECTION_MAX_HOSTS_ > CONNECTION_currentActiveHosts)
    return (CONNECTION_MAX_HOSTS_ - CONNECTION_currentActiveHosts)*5;
  else
    return 0;
}

/**
 * Consider switching the transport mechanism used for contacting
 * the given node. This function is called when the handler handles
 * an encrypted connection. For example, if we are sending SMTP
 * messages to a node behind a NAT box, but that node has established
 * a TCP connection to us, it might just be better to send replies
 * on that TCP connection instead of keeping SMTP going.<p>
 *
 * We can only successfully takeover if the transport is bidirectional
 * (can be associated). It also only makes sense if the cost is lower.
 * This method checks both.
 *
 * @param tsession the transport session that is for grabs
 * @param sender the identity of the other node
 **/
void considerTakeover(TSession * tsession,
		      HostIdentity * sender) {
  BufferEntry * be;

  ENTRY();
  if (tsession == NULL)
    return;
  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be != NULL) {
    if (be->status != STAT_DOWN) {
      unsigned int cost = -1;
      if (be->session.tsession != NULL)
	cost = transportGetCost(be->session.tsession->ttype);
      /* Question: doesn't this always do takeover in tcp/udp
	 case, which have the same costs? Should it? -IW 

	 Answer: this will always switch to TCP in the long run (if
	 that is possible) since udpAssociate always
	 returns SYSERR. This is intended since for long-running
	 sessions, TCP is the better choice. UDP is only better for
	 sending very few messages (e.g. attempting an initial exchange
	 to get to know each other). See also transport paper and the
	 data on throughput. - CG
      */
      if (transportGetCost(tsession->ttype) >= cost) {
	if (transportAssociate(tsession) == OK) {
	  transportDisconnect(be->session.tsession);
	  be->session.tsession = tsession;
	}
      } /* end if cheaper AND possible */
    } /* end if connected */
  }
  MUTEX_UNLOCK(&lock);
}


/**
 * Accept a session-key that has been sent by another host.
 * The other host must be known (public key)
 *
 * @param sender the identity of the sender host
 * @param tsession the transport session handle
 * @param msg message with the session key 
 **/
int acceptSessionKey(HostIdentity * sender,
		     TSession * tsession,
		     p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hostName;
  SESSIONKEY key;
  SKEY_Message * sessionkeySigned;

  ENTRY();
  statChange(stat_sessionkeys_received, 1);
  if (ntohs(msg->size) != sizeof(SKEY_Message))
    return SYSERR;
  sessionkeySigned = (SKEY_Message *) msg;
  IFLOG(LOG_DEBUG,
	hash2hex(&sender->hashPubKey,
		 &hostName));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: Received sessionkey from host %s.\n",
      &hostName);
#endif
  if (SYSERR == verifySKS(sender, 
			  sessionkeySigned)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: session key from %s failed verification, ignored!\n", 
	&hostName);
    return SYSERR;  /* rejected */
  }

  /* prepare file */
  if (sizeof(SESSIONKEY) != 
      decryptData(&sessionkeySigned->body.key,
		  &key,
		  sizeof(SESSIONKEY))) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: SKEY rejected from host %s\n",
	&hostName);
    return SYSERR;
  }

  MUTEX_LOCK(&lock);
  be = lookForHost(sender);
  if (be == NULL) 
    be = addHost(sender, NO);
  if (be == NULL) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_INFO, 
	"INFO: Session key exchange denied, slot busy.\n",
	&hostName);
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }

  if (be->created > (TIME_T)ntohl(sessionkeySigned->body.creationTime)) {
    LOG(LOG_INFO, 
	"INFO: key dropped, we've send or received a more recent key!\n");
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }

  /* if we have another connection established with the node,
     shut it down! */
  if (be->session.tsession != NULL) {
    LOG(LOG_DEBUG,
	"DEBUG: closing old connection, received new SKEY\n");
    shutdownConnection(be);
  }

  /* try to associate with an existing connection (if the other side
     contacted us with a bi-directional protocol, if that fails, establish
     our own connection */
  if (SYSERR == transportAssociate(tsession)) {
    HELO_Message * helo;

    tsession = NULL;
    if (SYSERR == identity2Helo(sender,
				ANY_PROTOCOL_NUMBER,
				NO,
				&helo)) {
      LOG(LOG_WARNING, 
	  "WARNING: sessionkey received, but could not find transport to reply\n");
      MUTEX_UNLOCK(&lock);
      return SYSERR;
    }
    if (SYSERR == transportConnect(helo,
				   &tsession)) {
      LOG(LOG_WARNING, 
	  "WARNING: sessionkey received, but transport failed to connect\n");
      FREE(helo);
      MUTEX_UNLOCK(&lock);
      return SYSERR;
    }
  }
  /* ok, everything set, let's change the state to waiting for ping
     and initialize the buffer entry */
  memcpy(&be->skey,
	 &key,
	 SESSIONKEY_LEN);
  be->session.tsession 
    = tsession;
  be->session.isEncrypted 
    = YES;
  be->session.mtu 
    = transportGetMTU(be->session.tsession->ttype);
  be->created 
    = ntohl(sessionkeySigned->body.creationTime);
  be->status 
    = STAT_WAITING_FOR_PONG;
  be->lastSequenceNumberReceived 
    = 0;
  be->lastPacketsBitmap 
    = (unsigned int) -1; /* all bits set */
  if (be->sendBuffer != NULL)
    LOG(LOG_FAILURE,
	"FAILURE: expected be->sendBuffer to be null in file %s:%d\n",
	__FILE__, __LINE__);
  be->lastSequenceNumberSend
    = 1;

#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: SKEY exchange - sending encrypted ping\n");
#endif
  checkAndPing(be);
  MUTEX_UNLOCK(&lock);
  return OK;
}

#define TRUSTDIR "data/credit/"

/**
 * Initialize this module.
 **/
void initConnection() {
  unsigned int i;
  char * gnHome;

  ENTRY();
  stat_sessionkeys_received 
    = statHandle("# sessionkeys received");
  stat_sessionkeys_verified 
    = statHandle("# valid sessionkeys received");
  stat_sessionkeys_transmitted
    = statHandle("# sessionkeys sent");
  stat_connections_shutdown
    = statHandle("# connections shutdown");
  stat_number_of_connections 
    = statHandle("# currently connected nodes");
  stat_number_of_bytes_noise_send
    = statHandle("# bytes noise sent");
  stat_number_of_bytes_send
    = statHandle("# encrypted bytes sent");
  stat_number_of_bytes_received
    = statHandle("# bytes decrypted");
  scl_nextHead 
    = NULL;
  scl_nextTail 
    = NULL;
  MUTEX_CREATE_RECURSIVE(&lock);
  max_bpm
    = 60 * getConfigurationInt("LOAD",
			       "MAXNETDOWNBPSTOTAL");
  if (max_bpm == 0)
    max_bpm = 50000 * 60;
  CONNECTION_currentActiveHosts 
    = 0;

  /* max-hosts is supposed to allow TARGET_MSG_SID MTU-sized messages
     per SECONDS_INACTIVE_DROP; maxbps=max_bpm/60 =>
     byte per SID = maxbpm*SID/60; divide by MTU to
     get number of messages that can be send per SID */
  CONNECTION_MAX_HOSTS_
    = max_bpm * SECONDS_INACTIVE_DROP / 60 / EXPECTED_MTU / TARGET_MSG_SID; 
  /* => for 50000 bps, we get 78 (rounded up to 128) connections! */
  
  if (CONNECTION_MAX_HOSTS_ < 4)
    CONNECTION_MAX_HOSTS_ = 4;
   i = 1;
  while (i< CONNECTION_MAX_HOSTS_)
    i*=2;
  CONNECTION_MAX_HOSTS_ = i; /* make sure it's a power of 2 */
  LOG(LOG_DEBUG,
      "DEBUG: connection goal is %d peers (%llu BPS bandwidth downstream)\n",
      i,
      max_bpm);

  CONNECTION_buffer_ 
    = (BufferEntry**) MALLOC(sizeof(BufferEntry*)*CONNECTION_MAX_HOSTS_);
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) 
    CONNECTION_buffer_[i] = NULL;

  registerp2pHandler(p2p_PROTO_SEQUENCE,
		     &checkSequenceNumber);
  registerp2pHandler(p2p_PROTO_HANGUP,
		     &handleHANGUP);
  registerp2pHandler(p2p_PROTO_CAPABILITY,
		     &handleCAPABILITY);
  registerCSHandler(CS_PROTO_CLIENT_COUNT,
		    &processGetConnectionCountRequest);
  addCronJob(&cronCountConnections,
	     1 * cronMINUTES,
	     30 * cronSECONDS,
	     NULL);
  addCronJob(&cronDecreaseLiveness,
	     1 * cronSECONDS, 
	     cronMINUTES / CONNECTION_MAX_HOSTS_ / 5,
	     NULL); 
  disable_random_padding = testConfigurationString("GNUNETD-EXPERIMENTAL",
						   "PADDING",
						   "NO");
  gnHome = getFileName("",
		       "GNUNET_HOME",
		       "Configuration file must specify a "
		       "directory for GNUnet to store "
		       "per-peer data under %s%s\n");
  trustDirectory = MALLOC(strlen(gnHome) + strlen(TRUSTDIR)+2);
  strcpy(trustDirectory, gnHome);
  FREE(gnHome);
  strcat(trustDirectory, "/");
  strcat(trustDirectory, TRUSTDIR);
  mkdirp(trustDirectory);
  addCronJob(&cronFlushTrustBuffer,
	     5 * cronMINUTES,
	     5 * cronMINUTES,
	     NULL); 
}


/**
 * Shutdown the connection module.
 **/
void doneConnection() {
  unsigned int i;
  BufferEntry * be;
  SendCallbackList * scl;
  
  ENTRY();
  delCronJob(&cronFlushTrustBuffer,
	     5 * cronMINUTES,
	     NULL);
  unregisterCSHandler(CS_PROTO_CLIENT_COUNT,
		      &processGetConnectionCountRequest);
  unregisterp2pHandler(p2p_PROTO_SEQUENCE,
		     &checkSequenceNumber);
  delCronJob(&cronCountConnections,
	     30 * cronSECONDS, 
	     NULL);
  delCronJob(&cronDecreaseLiveness,
	     cronMINUTES / CONNECTION_MAX_HOSTS_, 
	     NULL); 
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    BufferEntry * prev;

    prev = NULL;
    be = CONNECTION_buffer_[i];
    while (be != NULL) {
      LOG(LOG_DEBUG,
	  "DEBUG: closing connection: shutdown\n");
      shutdownConnection(be);
      flushHostCredit(be, NULL);    
      prev = be;
      be = be->overflowChain;
      FREE(prev);
    }
  } 
  MUTEX_DESTROY(&lock);
  FREE(CONNECTION_buffer_);
  FREE(trustDirectory);
  trustDirectory = NULL;
  CONNECTION_buffer_ = NULL;
  while (scl_nextHead != NULL) {
    scl = scl_nextHead;
    scl_nextHead = scl_nextHead->next;
    FREE(scl);
  }
  scl_nextTail = NULL;

}

/**
 * Increase the host credit by a value. 
 *
 * @param hostId is the identity of the host
 * @param value is the int value by which the host credit is to be increased or
 *        decreased
 * @returns the actual change in trust (positive or negative)
 */
unsigned int changeHostCredit(const HostIdentity * hostId, 
			      int value){
  BufferEntry * be;

  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be == NULL) {
    MUTEX_UNLOCK(&lock);
    return 0; /* not connected! */
  }
  if ( (be->trust & 0x7FFFFFF) + value < 0) {
    value = - be->trust;
    be->trust = 0x80000000;
  } else {
    be->trust += value;
    be->trust = be->trust | 0x80000000;
  }
  MUTEX_UNLOCK(&lock);
  return value;
}

/**
 * Obtain the trust record of a peer.
 * 
 * @param hostId the identity of the peer
 * @return the amount of trust we currently have in that peer
 **/
unsigned int getHostCredit(const HostIdentity * hostId) {
  BufferEntry * be;
  unsigned int trust;

  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be == NULL) {
    MUTEX_UNLOCK(&lock);
    return 0;
  }
  trust = be->trust;
  MUTEX_UNLOCK(&lock);
  return trust & 0x7FFFFFFF;
}

/**
 * Wrapper around forAllConnectedHosts.  Calls a given
 * method for each connected host.
 *
 * @param method method to call for each connected peer
 * @param arg second argument to method
 * @return number of connected nodes
 **/
int forEachConnectedNode(PerNodeCallback method,
			 void * arg) {
  fENHWrap wrap;
  int ret;

  wrap.method = method;
  wrap.arg = arg;
  MUTEX_LOCK(&lock);
  ret = forAllConnectedHosts(&fENHCallback,
			     &wrap);
  MUTEX_UNLOCK(&lock);
  return ret;
}

/**
 * Print the contents of the connection buffer (for debugging).
 **/
void printConnectionBuffer() {
  unsigned int i;
  BufferEntry * tmp;
  HexName hostName;
  HexName skey;
  unsigned short ttype;

  MUTEX_LOCK(&lock);
  ENTRY();
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    tmp = CONNECTION_buffer_[i];
    while (tmp != NULL) {
      if (tmp->status != STAT_DOWN) {
        IFLOG(LOG_MESSAGE,
  	      hash2hex(&tmp->session.sender.hashPubKey, 
  	  	       &hostName));
	IFLOG(LOG_MESSAGE,
	      hash2hex((HashCode160*) &tmp->skey,
		       &skey));
	ttype = 0;
	if (tmp->session.tsession != NULL)
	  ttype = tmp->session.tsession->ttype;
	LOG(LOG_MESSAGE,
  	    "CONNECTION-TABLE: %3d-%1d-%2d-%6u-%4ds"
	    " (of %ds) BPM %4ur %4ut %4ui-%3u: %20s-%16s\n",
	    i,
	    tmp->status,
	    ttype,
	    tmp->trust & 0x7FFFFFFF, 
	    (int) ((cronTime(NULL) - tmp->isAlive)/cronSECONDS), 
	    SECONDS_INACTIVE_DROP,
	    tmp->recently_received,
	    tmp->transmitted_limit,
	    tmp->idealized_limit,
	    tmp->sendBufferSize,
	    &hostName,
	    &skey);
      }
      tmp = tmp->overflowChain;
    }
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * Register a callback method that should be invoked whenever a
 * message is about to be send that has more than minimumPadding bytes
 * left before maxing out the MTU. The callback method can then be
 * used to add additional content to the message (instead of the
 * random noise that is added by otherwise). Note that if the MTU is 0
 * (for streams), the callback method will always be called with
 * padding set to the maximum number of bytes left in the buffer
 * allocated for the send.
 *
 * @param minimumPadding how large must the padding be in order
 *   to call this method?
 * @param callback the method to invoke. The receiver is the
 *   receiver of the message, position is the reference to the
 *   first unused position in the buffer where GNUnet is building
 *   the message, padding is the number of bytes left in that buffer.
 *   The callback method must return the number of bytes written to
 *   that buffer (must be a positive number).
 * @return OK if the handler was registered, SYSERR on error
 **/
int registerSendCallback(const unsigned int minimumPadding,
			 BufferFillCallback callback) {
  SendCallbackList * scl;

  ENTRY();
  scl = MALLOC(sizeof(SendCallbackList));
  scl->minimumPadding = minimumPadding;
  scl->callback = callback;
  scl->next = NULL;
  MUTEX_LOCK(&lock);
  if (scl_nextTail == NULL) {
    scl_nextHead = scl;
    scl_nextTail = scl;
  } else {
    scl_nextTail->next = scl;
    scl_nextTail = scl;
  }
  MUTEX_UNLOCK(&lock);
  return OK;  
}
 
/**
 * Unregister a handler that was registered with registerSendCallback.
 *
 * @param minimumPadding how large must the padding be in order
 *   to call this method?
 * @param callback the method to invoke. The receiver is the
 *   receiver of the message, position is the reference to the
 *   first unused position in the buffer where GNUnet is building
 *   the message, padding is the number of bytes left in that buffer.
 *   The callback method must return the number of bytes written to
 *   that buffer (must be a positive number).
 * @return OK if the handler was removed, SYSERR on error
 **/
int unregisterSendCallback(const unsigned int minimumPadding,
			   BufferFillCallback callback) {
  SendCallbackList * pos;
  SendCallbackList * prev;
  
  prev = NULL;
  MUTEX_LOCK(&lock);
  pos = scl_nextHead;
  while (pos != NULL) {
    if ( (pos->callback == callback) &&
	 (pos->minimumPadding == minimumPadding) ) {
      if (prev == NULL) 
	scl_nextHead = pos->next;
      else
	prev->next = pos->next;
      if (scl_nextTail == pos)
	scl_nextTail = prev;
      FREE(pos);
      MUTEX_UNLOCK(&lock);
      return OK;     
    }
    prev = pos;
    pos = pos->next;
  }
  MUTEX_UNLOCK(&lock);
  return SYSERR;
}

/**
 * We received a sign of life from this host. 
 * 
 * @param hostId the peer that gave a sign of live
 **/
void notifyPONG(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  ENTRY();
  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be != NULL) {
    switch (be->status) {
    case STAT_DOWN:
      break;
    case STAT_WAITING_FOR_PING:
      /* wrong message, ignore */
      break;
    case STAT_WAITING_FOR_PONG:
      be->status = STAT_UP;
      CONNECTION_currentActiveHosts++;
      cronTime(&be->isAlive);
      statChange(stat_number_of_connections,
		 1);
      IFLOG(LOG_DEBUG,
	    hash2hex(&hostId->hashPubKey, 
		     &hex));
#if DEBUG_CONNECTION
      LOG(LOG_DEBUG, 
	  "DEBUG: Marking host %s active.\n",
	  &hex);
#endif
      break;
    case STAT_UP:
      cronTime(&be->isAlive);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: undefined status (%d)!\n",
	  be->status);
      break;
    }
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * We received a sign of life from this host. 
 *
 * @param hostId the peer that send a PING.
 **/
void notifyPING(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  ENTRY();
  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be != NULL) {
    switch (be->status) {
    case STAT_DOWN:
      break;
    case STAT_WAITING_FOR_PONG:
      /* wrong message */
      break;
    case STAT_WAITING_FOR_PING:
      be->status = STAT_UP;
      CONNECTION_currentActiveHosts++;
      statChange(stat_number_of_connections,
		 1);
      cronTime(&be->isAlive);
      IFLOG(LOG_DEBUG,
	    hash2hex(&hostId->hashPubKey, 
		     &hex));
#if DEBUG_CONNECTION
      LOG(LOG_DEBUG, 
	  "DEBUG: Marking host %s active.\n",
	  &hex);
#endif
      break;
    case STAT_UP:
      cronTime(&be->isAlive);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: unknown status!\n");
      break;
    }
  }
  MUTEX_UNLOCK(&lock);  
}

/**
 * Send a message to all directly connected nodes.
 *
 * @param message the message to send
 * @param priority how important is the message? The higher, the more important
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void broadcast(p2p_HEADER * message,
	       unsigned int priority,
	       unsigned int maxdelay) {
  unsigned int i;
  BufferEntry * tmp;

  ENTRY();
  MUTEX_LOCK(&lock);
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    tmp = CONNECTION_buffer_[i];
    while (tmp != NULL) {
      /* we need no sync here as we only read,
         and concurrent rw access does not hurt */
      if (tmp->status == STAT_UP) 
	sendToNode(&tmp->session.sender,
		   message,
		   priority,
		   maxdelay);
      tmp = tmp->overflowChain;
    }
  }
  MUTEX_UNLOCK(&lock);    
}

/**
 * Send a message to a specific host (reply, enqueue)
 *
 * @param message the message to send (unencrypted!), first 2 bytes give size
 * @param hostId the identity of the receiver
 * @param priority how important is the message?
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void sendToNode(const HostIdentity * hostId,
		p2p_HEADER * message,
		unsigned int priority,
		unsigned int maxdelay) {
  BufferEntry * be;
#if DEBUG_CONNECTION
  HexName hex;

  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: sending message to host %s message of type %d\n",
      &hex,
      ntohs(message->requestType));
#endif
  ENTRY();
  MUTEX_LOCK(&lock);
  be = connectTo(hostId); 
  if ( (be != NULL) &&
       (be->status != STAT_DOWN) ) {  
    SendEntry * entry;
    unsigned short len = ntohs(message->size);

    entry = MALLOC(sizeof(SendEntry));
    entry->len = ntohs(message->size);
    entry->flags = SE_FLAG_NONE;
    entry->pri = priority;
    entry->transmissionTime = cronTime(NULL) + maxdelay;
    entry->callback = &copyCallback;
    entry->closure = MALLOC(len);
    memcpy(entry->closure,
	   message,
	   len);
    appendToBuffer(be,
		   entry);
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * Send an encrypted, on-demand build message to another node.
 *
 * @param hostId the target node
 * @param callback the callback to build the message
 * @param closure the second argument to callback
 * @param len how long is the message going to be?
 * @param importance how important is the message?
 * @param maxdelay how long can the message wait?
 **/
void unicast(const HostIdentity * hostId,
	     BuildMessageCallback callback,
	     void * closure,
	     unsigned short len,
	     unsigned int importance,
	     unsigned int maxdelay) {
  BufferEntry * be;
#if DEBUG_CONNECTION
  HexName hex;

  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: sending message to host %s message of size %d\n",
      &hex,
      len);
#endif
  ENTRY();
  MUTEX_LOCK(&lock);
  be = connectTo(hostId); 
  if ( (be != NULL) &&
       (be->status != STAT_DOWN) ) {  
    SendEntry * entry;

    entry = MALLOC(sizeof(SendEntry));
    entry->len = len;
    entry->flags = SE_FLAG_NONE;
    entry->pri = importance;
    entry->transmissionTime = cronTime(NULL) + maxdelay;
    entry->callback = callback;
    entry->closure = closure;
    appendToBuffer(be,
		   entry);
  }
  MUTEX_UNLOCK(&lock);
}

/**
 * Shutdown all connections (send HANGUPs, too).
 **/
void shutdownConnections() {
  MUTEX_LOCK(&lock);
  LOG(LOG_DEBUG,
      "DEBUG: shutdown of all connections\n");
  forAllConnectedHosts((BufferEntryCallback)&shutdownConnection, 
		       NULL);
  MUTEX_UNLOCK(&lock);
}

/**
 * Are we connected to this peer?
 *
 * @param hi the peer in question
 * @return NO if we are not connected, YES if we are
 **/
int isConnected(const HostIdentity * hi) {
  BufferEntry * be;

  MUTEX_LOCK(&lock);
  be = lookForHost(hi);
  MUTEX_UNLOCK(&lock);
  if (be == NULL) {
    return NO;
  } else {
    return YES;
  }
}
	       
/**
 * Decipher data coming in from a foreign host.
 *
 * @param data the data to decrypt
 * @param size the size of the encrypted data
 * @param hostId the sender host that encrypted the data 
 * @param result where to store the decrypted data, must
 *        be at least of size data->len long
 * @return the size of the decrypted data, SYSERR on error
 **/
int decryptFromHost(void * data,
		    const unsigned short size,
		    const HostIdentity * hostId,
		    void * result) {  
  BufferEntry * be;
  int res;
  HexName hex;

  ENTRY();
  statChange(stat_number_of_bytes_received, size);
  if ( (data == NULL) || 
       (hostId == NULL) ) {    
    LOG(LOG_FAILURE, 
	"CONNECTION: could not decrypt, message or hostId was NULL\n");
    BREAK();
    return SYSERR;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
#if DEBUG_CONNECTION
  LOG(LOG_DEBUG, 
      "DEBUG: decrypting message from host %s\n",
      &hex);
#endif
  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  if (be == NULL) {
    IFLOG(LOG_FAILURE,
	  hash2hex(&hostId->hashPubKey, 
		   &hex));
    LOG(LOG_FAILURE, 
	"FAILURE: decrypting message from host %s failed, no sessionkey!\n",
	&hex);
    MUTEX_UNLOCK(&lock);
    return SYSERR; /* could not decrypt */
  }
  res = decryptBlock(&be->skey, 
		     data,
		     size,
		     INITVALUE,
		     result);
  MUTEX_UNLOCK(&lock);
  return res;
}

/**
 * Compute the hashtable index of a host id.
 * 
 * @param hostId the ID of a peer
 * @return the index for this peer in the connection table
 **/
unsigned int computeIndex(const HostIdentity * hostId) {
  unsigned int res = (((unsigned int)hostId->hashPubKey.a) & 
		      ((unsigned int)(CONNECTION_MAX_HOSTS_ - 1)));
  if (res >= CONNECTION_MAX_HOSTS_)
    errexit("FATAL: CONNECTION_MAX_HOSTS_ not power of 2? (%d)\n",
	    CONNECTION_MAX_HOSTS_);
  return res;
}

/**
 * Obtain the lock for the connection module
 *
 * @return the lock
 **/
Mutex * getConnectionModuleLock() {
  return &lock;
}

/**
 * Notification for per-connection bandwidth tracking:
 * we received size bytes from hostId.  Note that only
 * encrypted messages are counted as "real" traffic.
 *
 * @param hostId the peer that send the message
 * @param size the size of the message
 **/
void trafficReceivedFrom(const HostIdentity * hostId,
			 const unsigned int size) {
  BufferEntry * be;

  ENTRY();
  MUTEX_LOCK(&lock);
  be = lookForHost(hostId);
  be->recently_received += size;
  MUTEX_UNLOCK(&lock);
}


/* end of connection.c */
